import os
import sys
from typing import Optional, Any, Callable, List, Dict, TYPE_CHECKING
import tomli
from pydantic.dataclasses import dataclass
from importlib import machinery
from chainlit.logger import logger

if TYPE_CHECKING:
    from chainlit.action import Action


# Get the directory the script is running from
root = os.getcwd()

config_dir = os.path.join(root, ".chainlit")
config_file = os.path.join(config_dir, "config.toml")

# Default config file created if none exists
DEFAULT_CONFIG_STR = """[project]
# Name of the app and chatbot.
name = "Chatbot"

# If true (default), the app will be available to anonymous users (once deployed).
# If false, users will need to authenticate and be part of the project to use the app.
public = true

# The project ID (found on https://cloud.chainlit.io).
# If provided, all the message data will be stored in the cloud.
# The project ID is required when public is set to false.
#id = ""

# Whether to enable telemetry (default: true). No personal data is collected.
enable_telemetry = true

# List of environment variables to be provided by each user to use the app.
user_env = []

# Hide the chain of thought details from the user in the UI.
hide_cot = false

# Link to your github repo. This will add a github button in the UI's header.
# github = ""

# Limit the number of requests per user.
#request_limit = "10 per day"
"""

# Set  and server URL
chainlit_prod_url = os.environ.get("CHAINLIT_PROD_URL")
chainlit_server = "https://cloud.chainlit.io"


@dataclass()
class ChainlitConfig:
    # Chainlit server URL. Used only for cloud features
    chainlit_server: str
    # Name of the app and chatbot. Used as the default message author.
    chatbot_name: str
    # Whether the app is available to anonymous users or only to team members.
    public: bool
    # Whether to enable telemetry. No personal data is collected.
    enable_telemetry: bool
    # List of environment variables to be provided by each user to use the app. If empty, no environment variables will be asked to the user.
    user_env: List[str]
    # Hide the chain of thought details from the user in the UI.
    hide_cot: bool
    # Path to the local langchain cache database
    lc_cache_path: str
    # Developer defined callbacks for each action. Key is the action name, value is the callback function.
    action_callbacks: Dict[str, Callable[["Action"], Any]]
    # Directory where the Chainlit project is located
    root = root
    # The url of the deployed app. Only set if the app is deployed.
    chainlit_prod_url = chainlit_prod_url
    # Link to your github repo. This will add a github button in the UI's header.
    github: Optional[str] = None
    # Limit the number of requests per user.
    request_limit: Optional[str] = None
    # Enables Cloud features if provided
    project_id: Optional[str] = None
    # Name of the module (python file) used in the run command
    module_name: Optional[str] = None
    # Module object loaded from the module_name
    module: Any = None
    # Bunch of callbacks defined by the developer
    on_stop: Optional[Callable[[], Any]] = None
    on_chat_start: Optional[Callable[[], Any]] = None
    on_message: Optional[Callable[[str], Any]] = None
    lc_run: Optional[Callable[[Any, str], str]] = None
    lc_postprocess: Optional[Callable[[Any], str]] = None
    lc_factory: Optional[Callable[[], Any]] = None
    lc_rename: Optional[Callable[[str], str]] = None


def init_config(log=False):
    """Initialize the configuration file if it doesn't exist."""
    if not os.path.exists(config_file):
        os.makedirs(config_dir, exist_ok=True)
        with open(config_file, "w", encoding="utf-8") as f:
            f.write(DEFAULT_CONFIG_STR)
            logger.info(f"Created default config file at {config_file}")
    elif log:
        logger.info(f"Config file already exists at {config_file}")


def reset_module_config():
    if not config:
        return

    module_fields = [
        "on_stop",
        "on_chat_start",
        "on_message",
        "lc_run",
        "lc_postprocess",
        "lc_factory",
        "lc_rename",
    ]

    for field in module_fields:
        setattr(config, field, None)


def load_module(target: str):
    """Load the specified module."""

    # Reset the config fields that belonged to the previous module
    reset_module_config()

    # Get the target's directory
    target_dir = os.path.dirname(os.path.abspath(target))

    # Add the target's directory to the Python path
    sys.path.insert(0, target_dir)

    loader = machinery.SourceFileLoader(target, target)
    config.module = loader.load_module()

    # Remove the target's directory from the Python path
    sys.path.pop(0)


def load_config():
    """Load the configuration from the config file."""
    init_config()
    with open(config_file, "rb") as f:
        toml_dict = tomli.load(f)
        # Load project settings
        project_settings = toml_dict.get("project", {})

        # If the developer did not explicitly opt out of telemetry, enable it
        enable_telemetry = project_settings.get("enable_telemetry")
        if enable_telemetry is None:
            enable_telemetry = True

        chatbot_name = project_settings.get("name")
        project_id = project_settings.get("id")
        public = project_settings.get("public")
        user_env = project_settings.get("user_env")
        hide_cot = project_settings.get("hide_cot", False)
        github = project_settings.get("github")
        request_limit = project_settings.get("request_limit", "")

        if not public and not project_id:
            raise ValueError("Project ID is required when public is set to false.")

        # Set cache path
        lc_cache_path = os.path.join(config_dir, ".langchain.db")

        config = ChainlitConfig(
            action_callbacks={},
            github=github,
            request_limit=request_limit,
            hide_cot=hide_cot,
            chainlit_server=chainlit_server,
            chatbot_name=chatbot_name,
            public=public,
            enable_telemetry=enable_telemetry,
            user_env=user_env,
            lc_cache_path=lc_cache_path,
            project_id=project_id,
        )

    return config


config = load_config()



import langchain
from typing import List, Optional
from langchain.llms import base as llm_base
from langchain.chat_models.base import BaseChatModel
from langchain.schema import (
    LLMResult,
    PromptValue,
)
from langchain.callbacks import get_callback_manager
from chainlit.lc.chainlit_handler import ChainlitCallbackHandler
from chainlit.lc.utils import get_llm_settings

# Monkey patch LangChain <= 0.0.153


def patched_generate(
    self: llm_base.BaseLLM, prompts: List[str], stop: Optional[List[str]] = None
) -> LLMResult:
    """Run the LLM on the given prompt and input."""
    # If string is passed in directly no errors will be raised but outputs will
    # not make sense.

    if not isinstance(prompts, list):
        raise ValueError(
            "Argument 'prompts' is expected to be of type List[str], received"
            f" argument of type {type(prompts)}."
        )
    # PATCH
    llm_settings = get_llm_settings(self, stop)
    disregard_cache = self.cache is not None and not self.cache
    if langchain.llm_cache is None or disregard_cache:
        # This happens when langchain.cache is None, but self.cache is True
        if self.cache is not None and self.cache:
            raise ValueError("Asked to cache, but no cache found at `langchain.cache`.")
        self.callback_manager.on_llm_start(
            {"name": self.__class__.__name__},
            prompts,
            verbose=self.verbose,
            llm_settings=llm_settings,
        )
        try:
            output = self._generate(prompts, stop=stop)
        except (KeyboardInterrupt, Exception) as e:
            self.callback_manager.on_llm_error(e, verbose=self.verbose)
            raise e
        self.callback_manager.on_llm_end(output, verbose=self.verbose)
        return output
    params = self.dict()
    params["stop"] = stop
    (
        existing_prompts,
        llm_string,
        missing_prompt_idxs,
        missing_prompts,
    ) = llm_base.get_prompts(params, prompts)

    if len(missing_prompts) > 0:
        self.callback_manager.on_llm_start(
            {"name": self.__class__.__name__},
            missing_prompts,
            verbose=self.verbose,
            llm_settings=llm_settings,
        )
        try:
            new_results = self._generate(missing_prompts, stop=stop)
        except (KeyboardInterrupt, Exception) as e:
            self.callback_manager.on_llm_error(e, verbose=self.verbose)
            raise e
        self.callback_manager.on_llm_end(new_results, verbose=self.verbose)
        llm_output = llm_base.update_cache(
            existing_prompts, llm_string, missing_prompt_idxs, new_results, prompts
        )
    else:
        # PATCH
        self.callback_manager.on_llm_start(
            {"name": self.__class__.__name__},
            prompts,
            verbose=self.verbose,
            llm_settings=llm_settings,
        )
        llm_output = {}
        self.callback_manager.on_llm_end(
            LLMResult(generations=[], llm_output=llm_output), verbose=self.verbose
        )
    generations = [existing_prompts[i] for i in range(len(prompts))]
    return LLMResult(generations=generations, llm_output=llm_output)


llm_base.BaseLLM.generate = patched_generate


def patched_generate_prompt(
    self, prompts: List[PromptValue], stop: Optional[List[str]] = None
) -> LLMResult:
    prompt_messages = [p.to_messages() for p in prompts]
    prompt_strings = [p.to_string() for p in prompts]

    # PATCH
    llm_settings = get_llm_settings(self, stop)
    self.callback_manager.on_llm_start(
        {"name": self.__class__.__name__},
        prompt_strings,
        verbose=self.verbose,
        llm_settings=llm_settings,
    )

    try:
        output = self.generate(prompt_messages, stop=stop)
    except (KeyboardInterrupt, Exception) as e:
        self.callback_manager.on_llm_error(e, verbose=self.verbose)
        raise e
    self.callback_manager.on_llm_end(output, verbose=self.verbose)
    return output


BaseChatModel.generate_prompt = patched_generate_prompt


get_callback_manager()._callback_manager.add_handler(ChainlitCallbackHandler())



from pydantic.dataclasses import dataclass
from dataclasses_json import dataclass_json
from chainlit.sdk import get_emit
from chainlit.telemetry import trace_event


@dataclass_json
@dataclass
class Action:
    name: str
    value: str
    description: str = ""
    forId: str = None

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")

    def send(self, for_id: str):
        emit = get_emit()
        if emit:
            trace_event(f"send {self.__class__.__name__}")
            self.forId = for_id
            emit("action", self.to_dict())

    def remove(self):
        emit = get_emit()
        if emit:
            trace_event(f"remove {self.__class__.__name__}")
            emit("remove_action", self.to_dict())



from typing import Any, Dict, List, Optional, Union
from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import AgentAction, AgentFinish, LLMResult
from chainlit.sdk import get_sdk, Chainlit
from chainlit.config import config
from chainlit.types import LLMSettings

IGNORE_LIST = ["AgentExecutor"]


class ChainlitCallbackHandler(BaseCallbackHandler):
    # Keep track of the formatted prompts to display them in the prompt playground.
    prompts_per_session: Dict[str, List[str]]
    # Keep track of the LLM settings for the last prompt
    llm_settings_per_session: Dict[str, LLMSettings]
    # Keep track of the call sequence, like [AgentExecutor, LLMMathChain, Calculator, ...]
    sequence_per_session: Dict[str, List[str]]
    # Keep track of the last prompt for each session
    last_prompt_per_session: Dict[str, Union[str, None]]
    # Keep track of whether the session is streaming or not
    stream_per_session: Dict[str, bool]

    # We want to handler to be called on every message
    always_verbose: bool = True

    def __init__(self) -> None:
        # Initialize dictionaries to store session data
        self.prompts_per_session = {}
        self.llm_settings_per_session = {}
        self.sequence_per_session = {}
        self.last_prompt_per_session = {}
        self.stream_per_session = {}

    def is_streaming(self) -> bool:
        sdk = get_sdk()
        if not sdk:
            return

        session_id = sdk.session["id"]
        return self.stream_per_session.get(session_id, False)

    def start_stream(self):
        sdk = get_sdk()
        if not sdk:
            return

        session_id = sdk.session["id"]
        author, indent, llm_settings = self.get_message_params(sdk)

        if author in IGNORE_LIST:
            return

        sdk.stream_start(author=author, indent=indent, llm_settings=llm_settings)
        self.stream_per_session[session_id] = True

    def send_token(self, token: str):
        sdk = get_sdk()
        if not sdk:
            return

        sdk.send_token(token)

    def end_stream(self):
        sdk = get_sdk()
        if not sdk:
            return

        session_id = sdk.session["id"]
        self.stream_per_session[session_id] = False

    def add_in_sequence(self, name: str):
        sdk = get_sdk()
        if not sdk:
            return

        session_id = sdk.session["id"]

        # Initialize session sequences if not already present
        if session_id not in self.sequence_per_session:
            self.sequence_per_session[session_id] = []

        sequence = self.sequence_per_session[session_id]

        sequence.append(name)

    def pop_sequence(self):
        sdk = get_sdk()
        if not sdk:
            return

        session_id = sdk.session["id"]

        # Remove the last element from the sequences
        if (
            session_id in self.sequence_per_session
            and self.sequence_per_session[session_id]
        ):
            self.sequence_per_session[session_id].pop()

    def add_prompt(self, prompt: str, llm_settings: LLMSettings = None):
        sdk = get_sdk()
        if not sdk:
            return

        session_id = sdk.session["id"]

        # Initialize session prompts if not already present
        if session_id not in self.prompts_per_session:
            self.prompts_per_session[session_id] = []

        self.prompts_per_session[session_id].append(prompt)

        if llm_settings:
            self.llm_settings_per_session[session_id] = llm_settings

    def pop_prompt(self):
        sdk = get_sdk()
        if not sdk:
            return

        session_id = sdk.session["id"]

        # Remove the last prompt from the session
        if (
            session_id in self.prompts_per_session
            and self.prompts_per_session[session_id]
        ):
            self.last_prompt_per_session[session_id] = self.prompts_per_session[
                session_id
            ].pop()

    def consume_last_prompt(self):
        sdk = get_sdk()
        if not sdk:
            return

        session_id = sdk.session["id"]

        last_prompt = self.last_prompt_per_session.get(session_id)
        self.last_prompt_per_session[session_id] = None
        return last_prompt

    def get_message_params(self, sdk: Chainlit):
        llm_settings = self.llm_settings_per_session.get(sdk.session["id"])

        sequence = self.sequence_per_session.get(sdk.session["id"])

        indent = len(sequence) if sequence else 0

        if sequence:
            if config.lc_rename:
                author = config.lc_rename(sequence[-1])
            else:
                author = sequence[-1]
        else:
            author = config.chatbot_name

        return author, indent, llm_settings

    def add_message(self, message, prompt: str = None, error=False):
        sdk = get_sdk()
        if not sdk:
            return

        author, indent, llm_settings = self.get_message_params(sdk)

        if author in IGNORE_LIST:
            return

        is_streaming = self.is_streaming()
        if is_streaming:
            self.end_stream()

        sdk.send_message(
            author=author,
            content=message,
            indent=indent,
            is_error=error,
            prompt=prompt,
            llm_settings=llm_settings,
            end_stream=is_streaming,
        )

    # Callbacks for various events

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        self.add_prompt(prompts[0], kwargs.get("llm_settings"))

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Do nothing."""
        if not self.is_streaming():
            self.start_stream()

        self.send_token(token)

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        """Do nothing."""
        self.pop_prompt()
        sdk = get_sdk()
        if not sdk:
            return
        if response.llm_output is not None:
            if "token_usage" in response.llm_output:
                token_usage = response.llm_output["token_usage"]
                if "total_tokens" in token_usage:
                    sdk.update_token_count(token_usage["total_tokens"])

    def on_llm_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        pass

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        self.add_in_sequence(serialized["name"])
        # Useful to display details button in the UI
        self.add_message("")

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        output_key = list(outputs.keys())[0]
        if output_key:
            prompt = self.consume_last_prompt()
            self.add_message(outputs[output_key], prompt)
        self.pop_sequence()

    def on_chain_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        self.add_message(str(error), error=True)
        self.pop_sequence()

    def on_tool_start(
        self, serialized: Dict[str, Any], inputs: Any, **kwargs: Any
    ) -> None:
        self.add_in_sequence(serialized["name"])
        self.add_message("")

    def on_tool_end(
        self,
        output: str,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        prompt = self.consume_last_prompt()
        self.add_message(output, prompt)
        self.pop_sequence()

    def on_tool_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        """Do nothing."""
        self.add_message(str(error), error=True)
        self.pop_sequence()

    def on_text(self, text: str, **kwargs: Any) -> None:
        pass

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        pass

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        """Run on agent end."""
        pass



# This is a simple example of a chainlit script.

from chainlit import send_message, ask_for_input, on_chat_start


@on_chat_start
def main():
    res = ask_for_input(content="What is your name?", timeout=30)
    if res:
        send_message(
            content=f"Your name is: {res['content']}.\nChainlit installation is working!",
        )



from pydantic.dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Dict
from abc import ABC, abstractmethod
from chainlit.sdk import get_sdk, BaseClient
from chainlit.telemetry import trace_event
from chainlit.types import ElementType, ElementDisplay, ElementSize


@dataclass_json
@dataclass
class Element(ABC):
    name: str
    type: ElementType
    display: ElementDisplay = "side"
    forId: str = None

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")

    @abstractmethod
    def persist(self, client: BaseClient, for_id: str = None) -> Dict:
        pass

    def before_emit(self, element: Dict) -> Dict:
        return element

    def send(self, for_id: str = None):
        sdk = get_sdk()

        element = None

        # Cloud is enabled, upload the element to S3
        if sdk.client:
            element = self.persist(sdk.client, for_id)

        if not element:
            element = self.to_dict()
            if for_id:
                element["forId"] = for_id

        if sdk.emit and element:
            trace_event(f"send {self.__class__.__name__}")
            element = self.before_emit(element)
            sdk.emit("element", element)


@dataclass
class LocalElement(Element):
    content: bytes = None

    def persist(self, client: BaseClient, for_id: str = None):
        if not self.content:
            raise ValueError("Must provide content")
        url = client.upload_element(content=self.content)
        if url:
            size = getattr(self, "size", None)
            language = getattr(self, "language", None)
            element = client.create_element(
                name=self.name,
                url=url,
                type=self.type,
                display=self.display,
                size=size,
                language=language,
                for_id=for_id,
            )
            return element


@dataclass
class RemoteElementBase:
    url: str


@dataclass
class ImageBase:
    type: ElementType = "image"
    size: ElementSize = "medium"


@dataclass
class RemoteElement(Element, RemoteElementBase):
    def persist(self, client: BaseClient, for_id: str = None):
        size = getattr(self, "size", None)
        language = getattr(self, "language", None)
        element = client.create_element(
            name=self.name,
            url=self.url,
            type=self.type,
            display=self.display,
            size=size,
            language=language,
            for_id=for_id,
        )
        return element


@dataclass
class LocalImage(ImageBase, LocalElement):
    """Useful to send an image living on the local filesystem to the UI."""

    path: str = None

    def __post_init__(self):
        if self.path:
            with open(self.path, "rb") as f:
                self.content = f.read()
        elif self.content:
            self.content = self.content
        else:
            raise ValueError("Must provide either path or content")


@dataclass
class RemoteImage(ImageBase, RemoteElement):
    """Useful to send an image based on an URL to the UI."""

    pass


@dataclass
class TextBase:
    text: str


@dataclass
class Text(LocalElement, TextBase):
    """Useful to send a text (not a message) to the UI."""

    type: ElementType = "text"
    content = bytes("", "utf-8")
    language: str = None

    def __post_init__(self):
        self.content = bytes(self.text, "utf-8")

    def before_emit(self, text_element):
        if "content" in text_element and isinstance(text_element["content"], bytes):
            text_element["content"] = text_element["content"].decode("utf-8")
        return text_element



import gevent
from gevent import monkey

monkey.patch_all()

from chainlit.lc import monkey

monkey.patch()

from chainlit.sdk import get_sdk
from chainlit.config import config
from chainlit.types import (
    LLMSettings,
    AskSpec,
    AskFileSpec,
    AskFileResponse,
    AskResponse,
)
from chainlit.telemetry import trace
from chainlit.version import __version__
from chainlit.logger import logger
from chainlit.server import socketio
from chainlit.action import Action
from chainlit.element import LocalImage, RemoteImage, Text
from chainlit.user_session import user_session
from typing import Callable, Any, List, Union
from dotenv import load_dotenv
import inspect
import os


env_found = load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))

if env_found:
    logger.info("Loaded .env file")


def wrap_user_function(user_function: Callable, with_task=False) -> Callable:
    """
    Wraps a user-defined function to accept arguments as a dictionary.

    Args:
        user_function (Callable): The user-defined function to wrap.

    Returns:
        Callable: The wrapped function.
    """

    def wrapper(*args):
        sdk = get_sdk()
        # Get the parameter names of the user-defined function
        user_function_params = list(inspect.signature(user_function).parameters.keys())

        # Create a dictionary of parameter names and their corresponding values from *args
        params_values = {
            param_name: arg for param_name, arg in zip(user_function_params, args)
        }

        if with_task and sdk:
            sdk.task_start()

        try:
            # Call the user-defined function with the arguments
            return user_function(**params_values)
        except Exception as e:
            logger.exception(e)
            if sdk:
                sdk.send_message(author="Error", is_error=True, content=str(e))
        finally:
            if with_task and sdk:
                sdk.task_end()

    return wrapper


@trace
def send_message(
    content: str,
    author=config.chatbot_name,
    prompt: str = None,
    language: str = None,
    indent=0,
    llm_settings: LLMSettings = None,
    end_stream=False,
    actions: List[Action] = [],
    elements: List[Union[LocalImage, RemoteImage, Text]] = [],
):
    """
    Send a message to the chatbot UI.
    If a project ID is configured, the message will be uploaded to the cloud storage.

    Args:
        content (str): The content of the message.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        prompt (str, optional): The prompt used to generate the message. If provided, enables the prompt playground for this message.
        language (str, optional): Language of the code is the content is code. See https://react-code-blocks-rajinwonderland.vercel.app/?path=/story/codeblock--supported-languages for a list of supported languages.
        indent (int, optional): If positive, the message will be nested in the UI.
        llm_settings (LLMSettings, optional): Settings of the LLM used to generate the prompt. This is useful for debug purposes in the prompt playground.
        end_stream (bool, optional): Pass True if this message was streamed.
        actions (List[Action], optional): A list of actions to send with the message.
        elements (List[Element], optional): A list of elements to send with the message.
    Returns:
        str: The message ID.
    """
    sdk = get_sdk()
    if sdk:
        msg_id = sdk.send_message(
            author=author,
            content=content,
            prompt=prompt,
            language=language,
            indent=indent,
            llm_settings=llm_settings,
            end_stream=end_stream,
        )

        for action in actions:
            action.send(for_id=msg_id)

        for element in elements:
            element.send(for_id=msg_id)

        return msg_id


@trace
def send_error_message(content: str, author=config.chatbot_name, indent=0):
    """
    Send an error message to the chatbot UI.
    If a project ID is configured, the message will be uploaded to the cloud storage.

    Args:
        content (str): The error message.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        indent (int, optional): If positive, the message will be nested in the UI.
    """
    sdk = get_sdk()
    if sdk:
        return sdk.send_message(
            author=author, content=content, is_error=True, indent=indent
        )


@trace
def ask_for_input(
    content: str, author=config.chatbot_name, timeout=60, raise_on_timeout=False
) -> Union[AskResponse, None]:
    """
    Ask for the user input before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    If a project ID is configured, the message will be uploaded to the cloud storage.

    Args:
        content (str): The content of the prompt.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        timeout (int, optional): The number of seconds to wait for an answer before raising a TimeoutError.
        raise_on_timeout (bool, optional): Whether to raise a socketio TimeoutError if the user does not answer in time.
    Returns:
        AskResponse: The response from the user include "msg" and "author" or None.
    """
    sdk = get_sdk()
    if sdk:
        spec = AskSpec(type="text", timeout=timeout)
        return sdk.send_ask_user(
            author=author, content=content, spec=spec, raise_on_timeout=raise_on_timeout
        )


@trace
def ask_for_file(
    title: str,
    accept: List[str],
    max_size_mb=2,
    author=config.chatbot_name,
    timeout=90,
    raise_on_timeout=False,
) -> Union[AskFileResponse, None]:
    """
    Ask the user to upload a file before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    If a project ID is configured, the file will be uploaded to the cloud storage.

    Args:
        title (str): Text displayed above the upload button.
        accept (List[str]): List of mime type to accept like ["text/csv", "application/pdf"]
        max_size_mb (int, optional): Maximum file size in MB.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        timeout (int, optional): The number of seconds to wait for an answer before raising a TimeoutError.
        raise_on_timeout (bool, optional): Whether to raise a socketio TimeoutError if the user does not answer in time.
    Returns:
        FileContent: The file content or None.
    """
    sdk = get_sdk()
    if sdk:
        spec = AskFileSpec(
            type="file", accept=accept, max_size_mb=max_size_mb, timeout=timeout
        )
        res = sdk.send_ask_user(
            author=author, content=title, spec=spec, raise_on_timeout=raise_on_timeout
        )
        if res:
            return AskFileResponse(**res)
        else:
            return None


@trace
def start_stream(
    author=config.chatbot_name,
    indent: int = 0,
    language: str = None,
    llm_settings: LLMSettings = None,
):
    """
    Start a streamed message.

    Args:
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        indent (int, optional): If positive, the message will be nested in the UI.
        language (str, optional): Language of the code is the content is code. See https://react-code-blocks-rajinwonderland.vercel.app/?path=/story/codeblock--supported-languages for a list of supported languages.
        llm_settings (LLMSettings, optional): Settings of the LLM used to generate the prompt. This is useful for debug purposes in the prompt playground.
    """
    sdk = get_sdk()
    if sdk:
        return sdk.stream_start(
            author=author, indent=indent, language=language, llm_settings=llm_settings
        )


@trace
def send_token(token: str):
    """
    Send a token belonging to the currently streamed message.

    Args:
        token (str): The token to send.
    """
    sdk = get_sdk()
    if sdk:
        return sdk.send_token(token)


@trace
def langchain_factory(func: Callable) -> Callable:
    """
    Plug and play decorator for the LangChain library.
    The decorated function should instantiate a new LangChain instance (Chain, Agent...).
    One instance per user session is created and cached.
    The per user instance is called every time a new message is received.

    Args:
        func (Callable[[], Any]): The factory function to create a new LangChain instance.

    Returns:
        Callable[[], Any]: The decorated factory function.
    """

    config.lc_factory = wrap_user_function(func, with_task=True)
    return func


@trace
def langchain_postprocess(func: Callable[[Any], str]) -> Callable:
    """
    Useful to post process the response a LangChain object instantiated with @langchain_factory.
    The decorated function takes the raw output of the LangChain object as input.
    The response will NOT be automatically sent to the UI, you need to call send_message.

    Args:
        func (Callable[[Any], str]): The post-processing function to apply after generating a response. Takes the response as parameter.

    Returns:
        Callable[[Any], str]: The decorated post-processing function.
    """

    config.lc_postprocess = wrap_user_function(func)
    return func


@trace
def on_message(func: Callable) -> Callable:
    """
    Framework agnostic decorator to react to messages coming from the UI.
    The decorated function is called every time a new message is received.

    Args:
        func (Callable[[str], Any]): The function to be called when a new message is received. Takes the input message.

    Returns:
        Callable[[str], Any]: The decorated on_message function.
    """

    config.on_message = wrap_user_function(func)
    return func


@trace
def langchain_run(func: Callable[[Any, str], str]) -> Callable:
    """
    Useful to override the default behavior of the LangChain object instantiated with @langchain_factory.
    Use when your agent run method has custom parameters.
    Takes the LangChain agent and the user input as parameters.
    The response will NOT be automatically sent to the UI, you need to call send_message.
    Args:
        func (Callable[[Any, str], str]): The function to be called when a new message is received. Takes the agent and user input as parameters and returns the output string.

    Returns:
        Callable[[Any, str], Any]: The decorated function.
    """

    config.lc_run = wrap_user_function(func)
    return func


@trace
def langchain_rename(func: Callable[[str], str]) -> Callable[[str], str]:
    """
    Useful to rename the LangChain tools/chains used in the agent and display more friendly author names in the UI.
    Args:
        func (Callable[[str], str]): The function to be called to rename a tool/chain. Takes the original tool/chain name as parameter.

    Returns:
        Callable[[Any, str], Any]: The decorated function.
    """

    config.lc_rename = wrap_user_function(func)
    return func


@trace
def on_chat_start(func: Callable) -> Callable:
    """
    Hook to react to the user websocket connection event.

    Args:
        func (Callable[], Any]): The connection hook to execute.

    Returns:
        Callable[], Any]: The decorated hook.
    """

    config.on_chat_start = wrap_user_function(func, with_task=True)
    return func


@trace
def on_stop(func: Callable) -> Callable:
    """
    Hook to react to the user stopping a conversation.

    Args:
        func (Callable[[], Any]): The stop hook to execute.

    Returns:
        Callable[[], Any]: The decorated stop hook.
    """

    config.on_stop = wrap_user_function(func)
    return func


def action_callback(name: str) -> Callable:
    """
    Callback to call when an action is clicked in the UI.

    Args:
        func (Callable[[Action], Any]): The action callback to exexute. First parameter is the action.
    """

    def decorator(func: Callable[[Action], Any]):
        config.action_callbacks[name] = wrap_user_function(func, with_task=True)
        return func

    return decorator


def sleep(duration: int):
    """
    Sleep for a given duration.
    Args:
        duration (int): The duration in seconds.
    """
    return socketio.sleep(duration)


__all__ = [
    "user_session",
    "Action",
    "LocalImage",
    "RemoteImage",
    "Text",
    "send_message",
    "send_error_message",
    "ask_for_input",
    "ask_for_file",
    "start_stream",
    "send_token",
    "langchain_factory",
    "langchain_postprocess",
    "langchain_run",
    "langchain_rename",
    "on_chat_start",
    "on_stop",
    "action",
    "sleep",
]



import os
from watchdog_gevent.observers import GeventObserver
from watchdog.events import FileSystemEventHandler
from chainlit.config import config, load_module
from chainlit.server import socketio
from chainlit.logger import logger

last_modified_time = 0
is_watching = False


class ChangeHandler(FileSystemEventHandler):
    def on_modified(self, event):
        global last_modified_time

        # Get the modified time of the file
        statbuf = os.stat(event.src_path)
        current_modified_time = statbuf.st_mtime

        file_ext = os.path.splitext(event.src_path)[1]

        if not file_ext in [".py", ".md"]:
            return

        # Check if the file was modified more than 0.5 seconds ago
        if (current_modified_time - last_modified_time) > 0.5:
            logger.info(f"event type: {event.event_type} path : {event.src_path}")

            # Load the module if the module name is specified in the config
            if config.module_name:
                load_module(config.module_name)

            # Emit a "reload" event to the socket
            socketio.emit("reload", {})

        last_modified_time = current_modified_time


def watch_directory():
    global is_watching

    # Return if already watching
    if is_watching:
        return

    is_watching = True
    event_handler = ChangeHandler()
    observer = GeventObserver()

    # Schedule the observer to watch the directory recursively
    observer.schedule(event_handler, config.root, recursive=True)

    # Start the observer
    observer.start()



import os
from chainlit.config import config
from chainlit.logger import logger


# Check if LangChain is installed and set up cache and callback handler
def patch():
    try:
        import langchain
        from langchain.cache import SQLiteCache

        if config.lc_cache_path:
            langchain.llm_cache = SQLiteCache(database_path=config.lc_cache_path)
            if not os.path.exists(config.lc_cache_path):
                logger.info(f"LangChain cache enabled: {config.lc_cache_path}")

        # New callback handler architecture
        if langchain.__version__ > "0.0.153":
            import chainlit.lc.new_monkey
        # Old callback handler architecture
        else:
            import chainlit.lc.old_monkey

        LANGCHAIN_INSTALLED = True
    except ImportError:
        LANGCHAIN_INSTALLED = False

    return LANGCHAIN_INSTALLED



import inspect
import langchain
from typing import List, Optional
from langchain.llms.base import BaseLLM, update_cache, get_prompts
from langchain.chat_models.base import BaseChatModel
from langchain.schema import (
    LLMResult,
)
from langchain.callbacks.manager import (
    CallbackManager,
    Callbacks,
)
from langchain.schema import (
    BaseMessage,
    LLMResult,
    get_buffer_string,
)
from chainlit.lc.utils import get_llm_settings
from chainlit.lc.chainlit_handler import ChainlitCallbackHandler

# Monkey patch LangChain > 0.0.153
chainlit_handler = ChainlitCallbackHandler()

orig_configure = CallbackManager.configure


def patched_configure(
    inheritable_callbacks: Callbacks = None,
    local_callbacks: Callbacks = None,
    verbose: bool = False,
):
    cbm = orig_configure(inheritable_callbacks, local_callbacks, verbose)
    cbm.add_handler(chainlit_handler, False)
    return cbm


CallbackManager.configure = patched_configure


def patched_generate(
    self,
    prompts: List[str],
    stop: Optional[List[str]] = None,
    callbacks: Callbacks = None,
) -> LLMResult:
    """Run the LLM on the given prompt and input."""
    # If string is passed in directly no errors will be raised but outputs will
    # not make sense.
    if not isinstance(prompts, list):
        raise ValueError(
            "Argument 'prompts' is expected to be of type List[str], received"
            f" argument of type {type(prompts)}."
        )
    # PATCH
    llm_settings = get_llm_settings(self, stop)
    disregard_cache = self.cache is not None and not self.cache
    callback_manager = CallbackManager.configure(
        callbacks, self.callbacks, self.verbose
    )
    new_arg_supported = inspect.signature(self._generate).parameters.get("run_manager")
    if langchain.llm_cache is None or disregard_cache:
        # This happens when langchain.cache is None, but self.cache is True
        if self.cache is not None and self.cache:
            raise ValueError("Asked to cache, but no cache found at `langchain.cache`.")
        # PATCH
        run_manager = callback_manager.on_llm_start(
            {"name": self.__class__.__name__}, prompts, llm_settings=llm_settings
        )
        try:
            output = (
                self._generate(prompts, stop=stop, run_manager=run_manager)
                if new_arg_supported
                else self._generate(prompts, stop=stop)
            )
        except (KeyboardInterrupt, Exception) as e:
            run_manager.on_llm_error(e)
            raise e
        run_manager.on_llm_end(output)
        return output
    params = self.dict()
    params["stop"] = stop
    (
        existing_prompts,
        llm_string,
        missing_prompt_idxs,
        missing_prompts,
    ) = get_prompts(params, prompts)
    if len(missing_prompts) > 0:
        # PATCH
        run_manager = callback_manager.on_llm_start(
            {"name": self.__class__.__name__},
            missing_prompts,
            llm_settings=llm_settings,
        )
        try:
            new_results = (
                self._generate(missing_prompts, stop=stop, run_manager=run_manager)
                if new_arg_supported
                else self._generate(missing_prompts, stop=stop)
            )
        except (KeyboardInterrupt, Exception) as e:
            run_manager.on_llm_error(e)
            raise e
        run_manager.on_llm_end(new_results)
        llm_output = update_cache(
            existing_prompts, llm_string, missing_prompt_idxs, new_results, prompts
        )
    else:
        # PATCH
        run_manager = callback_manager.on_llm_start(
            {"name": self.__class__.__name__},
            prompts,
            verbose=self.verbose,
            llm_settings=llm_settings,
        )
        llm_output = {}
        run_manager.on_llm_end(
            LLMResult(generations=[], llm_output=llm_output), verbose=self.verbose
        )

    generations = [existing_prompts[i] for i in range(len(prompts))]
    return LLMResult(generations=generations, llm_output=llm_output)


BaseLLM.generate = patched_generate


def patched_chat_generate(
    self,
    messages: List[List[BaseMessage]],
    stop: Optional[List[str]] = None,
    callbacks: Callbacks = None,
) -> LLMResult:
    """Top Level call"""

    # PATCH
    llm_settings = get_llm_settings(self, stop)

    callback_manager = CallbackManager.configure(
        callbacks, self.callbacks, self.verbose
    )
    message_strings = [get_buffer_string(m) for m in messages]

    # PATCH
    run_manager = callback_manager.on_llm_start(
        {"name": self.__class__.__name__}, message_strings, llm_settings=llm_settings
    )

    new_arg_supported = inspect.signature(self._generate).parameters.get("run_manager")
    try:
        results = [
            self._generate(m, stop=stop, run_manager=run_manager)
            if new_arg_supported
            else self._generate(m, stop=stop)
            for m in messages
        ]
    except (KeyboardInterrupt, Exception) as e:
        run_manager.on_llm_error(e)
        raise e
    llm_output = self._combine_llm_outputs([res.llm_output for res in results])
    generations = [res.generations for res in results]
    output = LLMResult(generations=generations, llm_output=llm_output)
    run_manager.on_llm_end(output)
    return output


BaseChatModel.generate = patched_chat_generate



from typing import Dict, Any, Optional
from python_graphql_client import GraphqlClient
from abc import ABC, abstractmethod
import uuid
import requests
from chainlit.types import ElementType, ElementSize
from chainlit.logger import logger


class BaseClient(ABC):
    project_id: str
    session_id: str

    @abstractmethod
    def create_conversation(self, session_id: str) -> int:
        pass

    @abstractmethod
    def create_message(self, variables: Dict[str, Any]) -> int:
        pass

    @abstractmethod
    def upload_element(self, content: bytes) -> int:
        pass

    @abstractmethod
    def create_element(
        self,
        type: ElementType,
        url: str,
        name: str,
        display: str,
        size: ElementSize = None,
        language: str = None,
        for_id: str = None,
    ) -> Dict[str, Any]:
        pass


class CloudClient(BaseClient):
    conversation_id: Optional[str] = None

    def __init__(self, project_id: str, session_id: str, access_token: str, url: str):
        self.project_id = project_id
        self.session_id = session_id
        self.url = url
        self.headers = {"Authorization": access_token}
        graphql_endpoint = f"{url}/api/graphql"
        self.client = GraphqlClient(endpoint=graphql_endpoint, headers=self.headers)

    def query(self, query: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Execute a GraphQL query.

        :param query: The GraphQL query string.
        :param variables: A dictionary of variables for the query.
        :return: The response data as a dictionary.
        """
        return self.client.execute(query=query, variables=variables)

    def check_for_errors(self, response: Dict[str, Any]):
        if "errors" in response:
            logger.error(response["errors"])
            return True
        return False

    def mutation(self, mutation: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Execute a GraphQL mutation.

        :param mutation: The GraphQL mutation string.
        :param variables: A dictionary of variables for the mutation.
        :return: The response data as a dictionary.
        """
        return self.client.execute(query=mutation, variables=variables)

    def create_conversation(self, session_id: str) -> int:
        mutation = """
        mutation ($projectId: String!, $sessionId: String!) {
            createConversation(projectId: $projectId, sessionId: $sessionId) {
                id
            }
        }
        """
        variables = {"projectId": self.project_id, "sessionId": session_id}
        res = self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not create conversation.")
            return None

        return int(res["data"]["createConversation"]["id"])

    def get_conversation_id(self):
        if not self.conversation_id:
            self.conversation_id = self.create_conversation(self.session_id)

        return self.conversation_id

    def create_message(self, variables: Dict[str, Any]) -> int:
        c_id = self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the message.")
            return None

        variables["conversationId"] = c_id

        mutation = """
        mutation ($conversationId: ID!, $author: String!, $content: String!, $language: String, $prompt: String, $llmSettings: Json, $isError: Boolean, $indent: Int, $authorIsUser: Boolean, $waitForAnswer: Boolean) {
            createMessage(conversationId: $conversationId, author: $author, content: $content, language: $language, prompt: $prompt, llmSettings: $llmSettings, isError: $isError, indent: $indent, authorIsUser: $authorIsUser, waitForAnswer: $waitForAnswer) {
                id
            }
        }
        """
        res = self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not persist message.")
            return None

        return int(res["data"]["createMessage"]["id"])

    def create_element(
        self,
        type: ElementType,
        url: str,
        name: str,
        display: str,
        size: ElementSize = None,
        language: str = None,
        for_id: str = None,
    ) -> Dict[str, Any]:
        c_id = self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the element.")
            return None

        mutation = """
        mutation ($conversationId: ID!, $type: String!, $url: String!, $name: String!, $display: String!, $size: String, $language: String, $forId: String) {
            createElement(conversationId: $conversationId, type: $type, url: $url, name: $name, display: $display, size: $size, language: $language, forId: $forId) {
                id,
                type,
                url,
                name,
                display,
                size,
                language,
                forId
            }
        }
        """
        variables = {
            "conversationId": c_id,
            "type": type,
            "url": url,
            "name": name,
            "display": display,
            "size": size,
            "language": language,
            "forId": for_id,
        }
        res = self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not persist element.")
            return None

        return res["data"]["createElement"]

    def upload_element(self, content: bytes) -> str:
        id = f"{uuid.uuid4()}"
        url = f"{self.url}/api/upload/file"
        body = {"projectId": self.project_id, "fileName": id}

        res = requests.post(url, json=body, headers=self.headers)

        if not res.ok:
            logger.error(f"Failed to upload file: {res.text}")
            return ""

        json_res = res.json()
        upload_details = json_res["post"]
        permanent_url = json_res["permanentUrl"]

        files = {"file": content}

        upload_response = requests.post(
            upload_details["url"], data=upload_details["fields"], files=files
        )

        if not upload_response.ok:
            logger.error(f"Failed to upload file: {res.text}")
            return ""

        url = f'{upload_details["url"]}/{upload_details["fields"]["key"]}'
        return permanent_url



from typing import Any
from chainlit.types import LLMSettings
from typing import List, Optional


def run_langchain_agent(agent: Any, input_str: str):
    if hasattr(agent, "input_keys"):
        input_key = agent.input_keys[0]
        raw_res = agent({input_key: input_str})
    else:
        raw_res = agent(input_str)

    if hasattr(agent, "output_keys"):
        output_key = agent.output_keys[0]
    else:
        output_key = None

    return raw_res, output_key


def get_llm_settings(llm, stop: Optional[List[str]] = None):
    if llm.__class__.__name__ == "OpenAI":
        return LLMSettings(
            model_name=llm.model_name,
            stop=stop,
            temperature=llm.temperature,
            max_tokens=llm.max_tokens,
            top_p=llm.top_p,
            frequency_penalty=llm.frequency_penalty,
            presence_penalty=llm.presence_penalty,
        )
    elif llm.__class__.__name__ == "ChatOpenAI":
        return LLMSettings(
            model_name=llm.model_name,
            stop=stop,
        )
    else:
        return None



from typing import Union
import time
import uuid
from chainlit.session import Session
from chainlit.types import LLMSettings, AskSpec
from chainlit.client import BaseClient
from socketio.exceptions import TimeoutError
import inspect


def current_milli_time():
    """Get the current time in milliseconds."""
    return round(time.time() * 1000)


class Chainlit:
    """
    Chainlit SDK class. The SDK is not directly exposed to the developer.
    Instead, the developer interacts with the SDK through the methods exposed in the __init__ file.
    """

    session: Session

    def __init__(self, session: Session) -> None:
        """Initialize Chainlit with the user session."""
        self.session = session

    def _get_session_property(self, property_name: str):
        """Helper method to get a property from the session."""
        if not hasattr(self, "session") or property_name not in self.session:
            return None
        return self.session[property_name]

    @property
    def emit(self):
        """Get the 'emit' property from the session."""
        return self._get_session_property("emit")

    @property
    def ask_user(self):
        """Get the 'ask_user' property from the session."""
        return self._get_session_property("ask_user")

    @property
    def client(self) -> Union[BaseClient, None]:
        """Get the 'client' property from the session."""
        return self._get_session_property("client")

    def send_message(
        self,
        author: str,
        content: str,
        prompt: str = None,
        language: str = None,
        indent=0,
        is_error=False,
        llm_settings: LLMSettings = None,
        end_stream=False,
    ):
        """Send a message to the UI."""
        if not self.emit:
            return

        if llm_settings is None and prompt is not None:
            llm_settings = LLMSettings()

        if llm_settings:
            llm_settings = llm_settings.to_dict()

        msg = {
            "author": author,
            "content": content,
            "indent": indent,
            "language": language,
            "isError": is_error,
            "prompt": prompt,
            "llmSettings": llm_settings,
        }
        if self.client:
            message_id = self.client.create_message(msg)
            msg["id"] = message_id

        if not "id" in msg:
            message_id = uuid.uuid4().hex
            msg["tempId"] = message_id

        msg["createdAt"] = current_milli_time()

        if end_stream:
            self.stream_end(msg)
        else:
            self.emit("message", msg)

        return str(message_id)

    def send_ask_timeout(self, author: str):
        """Send a prompt timeout message to the UI."""
        self.send_message(author=author, content="Time out", is_error=True)

        if self.emit:
            self.emit("ask_timeout", {})

    def clear_ask(self):
        """Clear the prompt from the UI."""
        if self.emit:
            self.emit("clear_ask", {})

    def send_ask_user(
        self, author: str, content: str, spec: AskSpec, raise_on_timeout=False
    ):
        """Send a prompt to the UI and wait for a response."""
        if not self.ask_user:
            return

        msg = {
            "author": author,
            "content": content,
            "waitForAnswer": True,
        }

        if self.client:
            message_id = self.client.create_message(msg)
            msg["id"] = message_id

        msg["createdAt"] = current_milli_time()

        try:
            # End the task temporarily so that the User can answer the prompt
            self.task_end()
            # Send the prompt to the UI
            res = self.ask_user({"msg": msg, "spec": spec.to_dict()}, spec.timeout)
            if self.client and res:
                # If cloud is enabled, store the response in the database/S3
                if spec.type == "text":
                    res_msg = {
                        "author": res["author"],
                        "authorIsUser": True,
                        "content": res["content"],
                    }
                    self.client.create_message(res_msg)
                elif spec.type == "file":
                    # TODO: upload file to S3
                    pass

            self.clear_ask()
            return res
        except TimeoutError as e:
            self.send_ask_timeout(author)
            if raise_on_timeout:
                raise e
        finally:
            self.task_start()

    def update_token_count(self, count: int):
        """Update the token count for the UI."""
        if not self.emit:
            return
        self.emit("token_usage", count)

    def task_start(self):
        """
        Send a task start signal to the UI.
        """
        if self.emit:
            self.emit("task_start", {})

    def task_end(self):
        """Send a task end signal to the UI."""
        if self.emit:
            self.emit("task_end", {})

    def stream_start(
        self,
        author: str,
        indent: int,
        language: str = None,
        llm_settings: LLMSettings = None,
    ):
        """Send a stream start signal to the UI."""
        if self.emit:
            self.emit(
                "stream_start",
                {
                    "author": author,
                    "indent": indent,
                    "language": language,
                    "content": "",
                    "llmSettings": llm_settings.to_dict() if llm_settings else None,
                },
            )

    def send_token(self, token: str):
        """Send a message token to the UI."""
        if self.emit:
            self.emit("stream_token", token)

    def stream_end(self, msg):
        """Send a stream end signal to the UI."""
        if self.emit:
            self.emit("stream_end", msg)


def get_sdk() -> Union[Chainlit, None]:
    """
    Get the Chainlit SDK instance from the current call stack.
    This unusual approach is necessary because:
     - we need to get the right SDK instance with the right websocket connection
     - to preserve a lean developer experience, we do not pass the SDK instance to every function call

    What happens is that we set __chainlit_sdk__ in the local variables when we receive a websocket message.
    Then we can retrieve it from the call stack when we need it, even if the developer's code has no idea about it.
    """
    attr = "__chainlit_sdk__"
    candidates = [i[0].f_locals.get(attr) for i in inspect.stack()]
    sdk = None
    for candidate in candidates:
        if candidate:
            sdk = candidate
            break
    return sdk


def get_emit():
    sdk = get_sdk()
    if sdk:
        return sdk.emit
    return None



import os
import json
from flask_cors import CORS
from flask import Flask, request, send_from_directory
from flask_socketio import SocketIO, ConnectionRefusedError
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address
from chainlit.config import config
from chainlit.lc.utils import run_langchain_agent
from chainlit.session import Session, sessions
from chainlit.user_session import user_sessions
from chainlit.client import CloudClient
from chainlit.sdk import Chainlit
from chainlit.markdown import get_markdown_str
from chainlit.action import Action
from chainlit.telemetry import trace, trace_event
from chainlit.logger import logger

root_dir = os.path.dirname(os.path.abspath(__file__))
build_dir = os.path.join(root_dir, "frontend/dist")

app = Flask(__name__, static_folder=build_dir)
CORS(app)
socketio = SocketIO(app, cors_allowed_origins="*", async_mode="gevent")

limiter = Limiter(
    get_remote_address,
    app=app,
    default_limits=[],
    storage_uri="memory://",
)


@app.route("/", defaults={"path": ""})
@app.route("/<path:path>")
def serve(path):
    """Serve the UI."""
    if path != "" and os.path.exists(app.static_folder + "/" + path):
        return send_from_directory(app.static_folder, path)
    else:
        return send_from_directory(app.static_folder, "index.html")


@app.route("/completion", methods=["POST"])
@limiter.limit(config.request_limit)
@trace
def completion():
    """Handle a completion request from the prompt playground."""

    import openai

    data = request.json
    llm_settings = data["settings"]
    user_env = data.get("userEnv", {})

    api_key = user_env.get("OPENAI_API_KEY", os.environ.get("OPENAI_API_KEY"))

    stop = llm_settings.pop("stop", None)
    model_name = llm_settings.pop("model_name", None)

    if model_name in ["gpt-3.5-turbo", "gpt-4"]:
        response = openai.ChatCompletion.create(
            api_key=api_key,
            model=model_name,
            messages=[{"role": "user", "content": data["prompt"]}],
            stop=stop,
            **llm_settings,
        )
        return response["choices"][0]["message"]["content"]
    else:
        response = openai.Completion.create(
            api_key=api_key,
            model=model_name,
            prompt=data["prompt"],
            stop=stop,
            **llm_settings,
        )
        return response["choices"][0]["text"]


@app.route("/project/settings", methods=["GET"])
def project_settings():
    """Return project settings. This is called by the UI before the establishing the websocket connection."""
    return {
        "public": config.public,
        "projectId": config.project_id,
        "chainlitServer": config.chainlit_server,
        "userEnv": config.user_env,
        "hideCot": config.hide_cot,
        "chainlitMd": get_markdown_str(config.root),
        "prod": bool(config.chainlit_prod_url),
        "appTitle": config.chatbot_name,
        "github": config.github,
    }


@socketio.on("connect")
def connect():
    """Handle socket connection."""
    session_id = request.sid
    client = None
    user_env = {}

    if config.user_env:
        # Check if requested user environment variables are provided
        if request.headers.get("user-env"):
            user_env = json.loads(request.headers.get("user-env"))
            for key in config.user_env:
                if key not in user_env:
                    trace_event("missing_user_env")
                    raise ConnectionRefusedError(
                        "Missing user environment variable: " + key
                    )

    access_token = request.headers.get("Authorization")
    if not config.public and not access_token:
        # Refuse connection if the app is private and no access token is provided
        trace_event("no_access_token")
        raise ConnectionRefusedError("No access token provided")
    elif access_token and config.project_id:
        # Create the cloud client
        client = CloudClient(
            project_id=config.project_id,
            session_id=session_id,
            access_token=access_token,
            url=config.chainlit_server,
        )

    # Function to send a message to this particular session
    def _emit(event, data):
        socketio.emit(event, data, to=session_id)

    # Function to ask the user a question
    def _ask_user(data, timeout):
        return socketio.call("ask", data, timeout=timeout, to=session_id)

    session = {
        "id": session_id,
        "emit": _emit,
        "ask_user": _ask_user,
        "client": client,
        "user_env": user_env,
    }  # type: Session
    sessions[session_id] = session

    if not config.lc_factory and not config.on_message and not config.on_chat_start:
        raise ValueError(
            "Module should at least expose one of @langchain_factory, @on_message or @on_chat_start function"
        )

    if config.lc_factory:

        def instantiate_agent(session):
            """Instantiate the langchain agent and store it in the session."""
            __chainlit_sdk__ = Chainlit(session)
            agent = config.lc_factory()
            session["agent"] = agent

        # Instantiate the agent in a background task since the connection is not yet accepted
        task = socketio.start_background_task(instantiate_agent, session)
        session["task"] = task

    if config.on_chat_start:

        def _on_chat_start(session):
            """Call the on_chat_start function provided by the developer."""
            __chainlit_sdk__ = Chainlit(session)
            config.on_chat_start()

        # Send the ask in a backgroudn task since the connection is not yet accepted
        task = socketio.start_background_task(_on_chat_start, session)
        session["task"] = task

    trace_event("connection_successful")


@socketio.on("disconnect")
def disconnect():
    """Handle socket disconnection."""

    if request.sid in sessions:
        # Clean up the session
        session = sessions.pop(request.sid)
        task = session.get("task")
        if task:
            # If a background task is running, kill it
            task.kill()

    if request.sid in user_sessions:
        # Clean up the user session
        user_sessions.pop(request.sid)


@socketio.on("stop")
def stop():
    """Handle a stop request from the client."""
    trace_event("stop_task")
    session = sessions.get(request.sid)
    if not session:
        return

    task = session.get("task")

    if task:
        task.kill()
        session["task"] = None

        __chainlit_sdk__ = Chainlit(session)

        if config.on_stop:
            config.on_stop()

        __chainlit_sdk__.send_message(
            author="System", content="Conversation stopped by the user."
        )


def need_session(id: str):
    """Return the session with the given id."""

    session = sessions.get(id)
    if not session:
        raise ValueError("Session not found")
    return session


def process_message(session: Session, author: str, input_str: str):
    """Process a message from the user."""

    __chainlit_sdk__ = Chainlit(session)
    try:
        __chainlit_sdk__.task_start()

        if session["client"]:
            # If cloud is enabled, persist the message
            session["client"].create_message(
                {
                    "author": author,
                    "content": input_str,
                    "authorIsUser": True,
                }
            )

        langchain_agent = session.get("agent")
        if langchain_agent:
            # If a langchain agent is available, run it
            if config.lc_run:
                # If the developer provided a custom run function, use it
                config.lc_run(langchain_agent, input_str)
                return
            else:
                # Otherwise, use the default run function
                raw_res, output_key = run_langchain_agent(langchain_agent, input_str)

                if config.lc_postprocess:
                    # If the developer provided a custom postprocess function, use it
                    config.lc_postprocess(raw_res)
                    return
                elif output_key is not None:
                    # Use the output key if provided
                    res = raw_res[output_key]
                else:
                    # Otherwise, use the raw response
                    res = raw_res
            # Finally, send the response to the user
            __chainlit_sdk__.send_message(author=config.chatbot_name, content=res)

        elif config.on_message:
            # If no langchain agent is available, call the on_message function provided by the developer
            config.on_message(input_str)
    except Exception as e:
        logger.exception(e)
        __chainlit_sdk__.send_message(author="Error", is_error=True, content=str(e))
    finally:
        __chainlit_sdk__.task_end()


@app.route("/message", methods=["POST"])
@limiter.limit(config.request_limit)
@trace
def message():
    """Handle a message from the UI."""

    body = request.json
    session_id = body["sessionId"]
    input_str = body["content"].strip()
    author = body["author"]

    session = need_session(session_id)

    task = socketio.start_background_task(process_message, session, author, input_str)
    session["task"] = task
    task.join()
    session["task"] = None

    return {"success": True}


def process_action(session: Session, action: Action):
    __chainlit_sdk__ = Chainlit(session)
    callback = config.action_callbacks.get(action.name)
    if callback:
        callback(action)
    else:
        logger.warning("No callback found for action %s", action.name)


@app.route("/action", methods=["POST"])
@limiter.limit(config.request_limit)
@trace
def on_action():
    """Handle an action call from the UI."""

    body = request.json
    session_id = body["sessionId"]
    action = Action(**body["action"])

    session = need_session(session_id)

    task = socketio.start_background_task(process_action, session, action)
    session["task"] = task
    task.join()
    session["task"] = None

    return {"success": True}



import click
import os
import sys
import webbrowser
from chainlit.config import config, init_config, load_module
from chainlit.watch import watch_directory
from chainlit.markdown import init_markdown
from chainlit.cli.auth import login, logout
from chainlit.cli.deploy import deploy
from chainlit.cli.utils import check_file
from chainlit.telemetry import trace_event
from chainlit.logger import logger

# Set the default port for the server
PORT = 8000


# Create the main command group for Chainlit CLI
@click.group(context_settings={"auto_envvar_prefix": "CHAINLIT"})
@click.version_option(prog_name="Chainlit")
def cli():
    return


# Define the function to run Chainlit with provided options
def run_chainlit(
    target: str, watch=False, headless=False, debug=False, args=None, **kwargs
):
    check_file(target)
    # Load the module provided by the user
    config.module_name = target
    load_module(config.module_name)

    # Create the chainlit.md file if it doesn't exist
    init_markdown(config.root)

    # Enable file watching if the user specified it
    if watch:
        watch_directory()

    from chainlit.server import socketio, app

    # Open the browser if in development mode
    def open_browser(headless: bool):
        if not headless:
            # Wait two seconds to allow the server to start
            socketio.sleep(2)
            url = f"http://127.0.0.1:{PORT}"
            logger.info(f"Your app is available at {url}")
            webbrowser.open(url)

    socketio.start_background_task(open_browser, headless)
    # Start the server
    socketio.run(app, host="0.0.0.0", port=PORT, debug=debug, use_reloader=False)


# Define the "run" command for Chainlit CLI
@cli.command("run")
@click.argument("target", required=True, envvar="CHAINLIT_RUN_TARGET")
@click.option("-w", "--watch", default=False, is_flag=True, envvar="CHAINLIT_WATCH")
@click.option(
    "-h", "--headless", default=False, is_flag=True, envvar="CHAINLIT_HEADLESS"
)
@click.option("-d", "--debug", default=False, is_flag=True, envvar="CHAINLIT_DEBUG")
@click.option("-c", "--ci", default=False, is_flag=True, envvar="CHAINLIT_CI")
@click.argument("args", nargs=-1)
def chainlit_run(target, watch, headless, debug, ci, args=None, **kwargs):
    if ci:
        logger.info("Running in CI mode")
        config.enable_telemetry = False
    else:
        trace_event("chainlit run")

    run_chainlit(target, watch, headless, debug, args, **kwargs)


@cli.command("deploy")
@click.argument("target", required=True, envvar="CHAINLIT_RUN_TARGET")
@click.argument("args", nargs=-1)
def chainlit_deploy(target, args=None, **kwargs):
    trace_event("chainlit deploy")
    raise NotImplementedError("Deploy is not yet implemented")
    deploy(target)


@cli.command("hello")
@click.argument("args", nargs=-1)
def chainlit_hello(args=None, **kwargs):
    trace_event("chainlit hello")
    dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    hello_path = os.path.join(dir_path, "hello.py")
    run_chainlit(hello_path)


@cli.command("login")
@click.argument("args", nargs=-1)
def chainlit_login(args=None, **kwargs):
    trace_event("chainlit login")
    login()
    sys.exit(0)


@cli.command("logout")
@click.argument("args", nargs=-1)
def chainlit_logout(args=None, **kwargs):
    trace_event("chainlit logout")
    logout()
    sys.exit(0)


@cli.command("init")
@click.argument("args", nargs=-1)
def chainlit_init(args=None, **kwargs):
    trace_event("chainlit init")
    init_config(log=True)
